package eu.jonahbauer.chat.bot.impl;

import eu.jonahbauer.chat.bot.api.ChatBot;
import eu.jonahbauer.chat.bot.config.BotConfig;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.TestOnly;

import java.util.Objects;
import java.util.ServiceLoader;
import java.util.Set;
import java.util.function.Supplier;
import java.util.stream.Collectors;

public final class ChatBotFactory<T extends ChatBot> {
    public static final ScopedValue<BotConfig> BOT_CONFIG = ScopedValue.newInstance();
    private static final ServiceLoader<ChatBot> SERVICE_LOADER = ServiceLoader.load(ChatBot.class);

    private final @NotNull Class<T> type;
    private final @NotNull Supplier<T> delegate;

    public static @NotNull Set<@NotNull Class<? extends ChatBot>> implementations() {
        return SERVICE_LOADER.stream().map(ServiceLoader.Provider::type).collect(Collectors.toUnmodifiableSet());
    }

    @SuppressWarnings("unchecked")
    public ChatBotFactory(@NotNull Class<T> type) {
        if (!ChatBot.class.isAssignableFrom(type)) throw new IllegalArgumentException();

        this.type = type;
        this.delegate = (Supplier<T>) SERVICE_LOADER.stream()
                .filter(p -> p.type().equals(type)).findFirst()
                .orElseThrow(() -> new BotCreationException("No suitable provider found: " + type.getName()));
    }

    @TestOnly
    ChatBotFactory(@NotNull Class<T> type, @NotNull Supplier<T> supplier) {
        if (!ChatBot.class.isAssignableFrom(type)) throw new IllegalArgumentException();
        this.type = type;
        this.delegate = Objects.requireNonNull(supplier, "supplier");
    }

    public @NotNull Class<T> type() {
        return type;
    }

    public @NotNull T create() {
        return create(BotConfig.EMPTY);
    }

    public @NotNull T create(@NotNull BotConfig config) {
        try {
            return ScopedValue.where(BOT_CONFIG, config).get(delegate);
        } catch (Throwable t) {
            throw new BotCreationException("Exception in constructor: " + type.getName(), t);
        }
    }
}
