package eu.jonahbauer.chat.bot.database.impl;

import eu.jonahbauer.chat.bot.database.Entity;
import eu.jonahbauer.chat.bot.database.annotations.Column;
import eu.jonahbauer.chat.bot.database.annotations.Table;
import lombok.Lombok;
import lombok.SneakyThrows;
import lombok.experimental.UtilityClass;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.reflect.RecordComponent;
import java.util.HashMap;
import java.util.Map;

import static java.lang.invoke.MethodHandles.*;
import static java.lang.invoke.MethodType.methodType;

@UtilityClass
public class Helper {
    public static @NotNull String getQuotedColumnName(@NotNull RecordComponent component) {
        return quote(getColumnName(component));
    }
    public static @NotNull String getColumnName(@NotNull RecordComponent component) {
        return getColumnName(component.getDeclaringRecord(), component.getName());
    }
    public static @NotNull String getColumnName(@NotNull Class<?> type, @NotNull String component) {
        var out = columnNameCache.get(type).get(component);
        if (out == null) throw new IllegalArgumentException("invalid component name " + component);
        return out;
    }
    private static final @NotNull ClassValue<@NotNull Map<@NotNull String, @NotNull String>> columnNameCache = new ClassValue<>() {
        @Override
        protected @NotNull Map<@NotNull String, @NotNull String> computeValue(@NotNull Class<?> type) {
            var out = new HashMap<String, String>();

            for (var component : type.getRecordComponents()) {
                var column = component.getAnnotation(Column.class);
                if (column != null) {
                    out.put(component.getName(), column.value());
                } else {
                    out.put(component.getName(), getDatabaseName(component.getName()));
                }
            }

            return Map.copyOf(out);
        }
    };

    public static @NotNull String getQuotedTableName(@NotNull Class<? extends Record> type) {
        return tableNameCache.get(type);
    }
    private static final @NotNull ClassValue<@NotNull String> tableNameCache = new ClassValue<>() {
        @Override
        protected @NotNull String computeValue(@NotNull Class<?> type) {
            var table = type.getAnnotation(Table.class);
            if (table != null) {
                return quote(table.value());
            } else {
                return quote(getDatabaseName(type.getSimpleName()));
            }
        }
    };

    public static boolean isIdColumn(@NotNull RecordComponent component) {
        return Entity.class.isAssignableFrom(component.getDeclaringRecord())
               && component.getType() == long.class
               && "id".equals(component.getName());
    }

    @SuppressWarnings("unchecked")
    public static <T extends Record & Entity<T>> WithId<T> getWithId(@NotNull Class<T> type) {
        return (WithId<T>) withIdCache.get(type);
    }
    private static final @NotNull ClassValue<@NotNull WithId<?>> withIdCache = new ClassValue<>() {
        @Override
        @SneakyThrows
        protected WithId<?> computeValue(Class<?> type) {
            if (!type.isRecord()) throw new IllegalArgumentException();
            if (!Entity.class.isAssignableFrom(type)) throw new IllegalArgumentException();

            var lookup = MethodHandles.lookup();

            var components = type.getRecordComponents();
            var types = new Class<?>[components.length];
            var accessors = new MethodHandle[components.length];
            var reorder = new int[components.length];

            for (int i = 0; i < components.length; i++) {
                var component = components[i];
                types[i] = component.getType();
                if (isIdColumn(component)) {
                    accessors[i] = identity(long.class);
                    reorder[i] = 1;
                } else {
                    accessors[i] = lookup.unreflect(component.getAccessor());
                    reorder[i] = 0;
                }
            }

            var ctor = lookup.findConstructor(type, methodType(void.class, types));
            var cloner = filterArguments(ctor, 0, accessors);
            var wither = permuteArguments(cloner, methodType(type, type, long.class), reorder);

            return create(wither.asType(methodType(Record.class, Record.class, long.class)));
        }

        @SuppressWarnings("unchecked")
        private <T extends Record & Entity<T>> WithId<T> create(@NotNull MethodHandle handle) {
            return (entity, id) -> {
                try {
                    return (T) handle.invokeExact(entity, id);
                } catch (Throwable t) {
                    throw Lombok.sneakyThrow(t);
                }
            };
        }
    };


    private static final @NotNull Logger log = LoggerFactory.getLogger("eu.jonahbauer.chat.bot.database.SQL");
    public static void logSqlStatement(@NotNull String sql) {
        log.debug(sql);
    }

    private static @NotNull String getDatabaseName(@NotNull String name) {
        if (name.length() <= 1) return name;

        StringBuilder out = null;
        for (int i = 1, length = name.length(); i < length; i++) {
            var prev = name.charAt(i - 1);
            var curr = name.charAt(i);
            if (Character.isLowerCase(prev) && Character.isUpperCase(curr)) {
                if (out == null) {
                    out = new StringBuilder(name.length() + 1);
                    out.append(name, 0, i);
                    out.setCharAt(0, Character.toLowerCase(out.charAt(0)));
                }
                out.append('_');
                out.append(Character.toLowerCase(curr));
            } else if (out != null) {
                out.append(curr);
            }
        }

        if (out == null) {
            if (Character.isUpperCase(name.charAt(0))) {
                return Character.toLowerCase(name.charAt(0)) + name.substring(1);
            } else {
                return name;
            }
        } else {
            return out.toString();
        }
    }

    public static @NotNull String quote(@NotNull String name) {
        var out = new StringBuilder(name.length() + 2);
        out.append('"');
        for (int i = 0, length = name.length(); i < length; i++) {
            var chr = name.charAt(i);
            if (chr == '"') out.append('"');
            if (chr == '\n') out.append('\\');

            out.append(chr);
        }
        out.append('"');
        return out.toString();
    }
}
