package eu.jonahbauer.chat.bot.pizza;

import eu.jonahbauer.chat.bot.database.Database;
import eu.jonahbauer.chat.bot.database.NonUniqueResultException;
import eu.jonahbauer.chat.bot.pizza.model.*;
import eu.jonahbauer.chat.bot.pizza.model.table.*;
import eu.jonahbauer.chat.bot.pizza.model.view.OrderDetail;
import eu.jonahbauer.chat.bot.pizza.model.view.OrderItemDetail;
import eu.jonahbauer.chat.bot.pizza.util.Pair;
import org.intellij.lang.annotations.Language;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.sql.SQLException;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

public class PizzaService {
    private final @NotNull Database database;

    public PizzaService(@NotNull Database database) {
        this.database = Objects.requireNonNull(database);
    }

    public @NotNull OrderDetail getCurrentOrder() throws SQLException {
        @Language("SQL")
        var query = "SELECT * FROM order_detail WHERE active = TRUE ORDER BY deadline DESC LIMIT 1";
        return database.executeUniqueQuery(OrderDetail.class, query)
                .orElseThrow(() -> new PizzaBotException("Aktuell läuft keine Pizzabestellung."));
    }

    public @NotNull Pair<@NotNull Restaurant, @NotNull List<@NotNull Pizza>> getRestaurantWithPizzas(long restaurantId) throws SQLException {
        return database.transaction(() -> {
            var restaurant = database.findById(Restaurant.class, restaurantId).orElseThrow();
            var pizzas = database.findAllWhere(Pizza.class, Pizza.Fields.restaurantId, restaurantId);
            return Pair.of(restaurant, pizzas);
        });
    }

    // Payment

    public void save(@NotNull Order order) throws SQLException {
        database.update(order);
    }

    public @NotNull User getUserByName(@NotNull String name) throws SQLException {
        try {
            @Language("SQL")
            var query = """
                    SELECT "user".* FROM "user"
                    WHERE "user"."username" LIKE ?
                    """;
            return database.executeUniqueQuery(User.class, query, stmt -> stmt.setString(1, name + "%"))
                    .orElseThrow(() -> new PizzaBotException("Der Benutzer existiert nicht."));
        } catch (NonUniqueResultException _) {
            throw new PizzaBotException("Der Benutzer ist nicht eindeutig.");
        }
    }

    public @NotNull User getUserByEventAndName(long eventId, @NotNull String name) throws SQLException {
        try {
            @Language("SQL")
            var query = """
                    SELECT "user".* FROM "user"
                    WHERE "user"."username" LIKE ?
                        AND EXISTS(
                            SELECT * FROM "order_item_to_user"
                                JOIN "order_item" ON "order_item_to_user"."order_item_id" = "order_item"."id"
                                JOIN "order" ON "order_item"."order_id" = "order"."id"
                            WHERE "order_item_to_user"."user_id" = "user"."id"
                                AND "order"."event_id" = ?
                        )
                    """;
            return database.executeUniqueQuery(User.class, query, stmt -> {
                stmt.setString(1, name + "%");
                stmt.setLong(2, eventId);
            }).orElseThrow(() -> new PizzaBotException("Der Benutzer existiert nicht oder hat an keiner Pizzabestellung teilgenommen."));
        } catch (NonUniqueResultException _) {
            throw new PizzaBotException("Der Benutzer ist nicht eindeutig.");
        }
    }

    public void paymentByItem(long itemId, boolean payed) throws SQLException {
        database.executeUpdate("UPDATE order_item_to_user SET payed = ? WHERE order_item_id = ?", stmt -> {
            stmt.setBoolean(1, payed);
            stmt.setLong(2, itemId);
        });
    }

    public int paymentByEventAndUser(long eventId, long userId, boolean payed) throws SQLException {
        @Language("SQL")
        var query = """
                  UPDATE "order_item_to_user"
                  SET "payed" = ?
                  WHERE "order_item_to_user"."user_id" = ?
                      AND EXISTS(
                          SELECT *
                          FROM "order_item"
                              JOIN "order" ON "order_item"."order_id" = "order"."id"
                          WHERE "order"."event_id" = ? AND "order_item"."id" = "order_item_to_user"."order_item_id"
                      )
                  """;
        return database.executeUpdate(query, stmt -> {
            stmt.setBoolean(1, payed);
            stmt.setLong(2, userId);
            stmt.setLong(3, eventId);
        });
    }

    public void paymentByItemAndUser(long itemId, long userId, boolean payed) throws SQLException {
        database.executeUpdate("UPDATE order_item_to_user SET payed = ? WHERE order_item_id = ? AND user_id = ?", stmt -> {
            stmt.setBoolean(1, payed);
            stmt.setLong(2, itemId);
            stmt.setLong(3, userId);
        });
    }

    // Order

    public @NotNull Optional<Pizza> getPizzaById(long id) throws SQLException {
        return database.findById(Pizza.class, id);
    }

    public @NotNull Pizza getPizza(long restaurantId, @NotNull String nameOrNummer) throws SQLException {
        try {
            var nummer = Integer.parseInt(nameOrNummer);
            return getPizzaByNummer(restaurantId, nummer);
        } catch (NumberFormatException _) {
            return getPizzaByName(restaurantId, nameOrNummer);
        }
    }

    private @NotNull Pizza getPizzaByNummer(long restaurantId, int nummer) throws SQLException {
        @Language("SQL")
        var query = "SELECT * FROM pizza WHERE restaurant_id = ? AND number = ?";
        return database.executeUniqueQuery(Pizza.class, query, stmt -> {
            stmt.setLong(1, restaurantId);
            stmt.setInt(2, nummer);
        }).orElseThrow(() -> new PizzaBotException("Diese Pizza kenne ich nicht."));
    }

    private @NotNull Pizza getPizzaByName(long restaurantId, @NotNull String name) throws SQLException {
        try {
            @Language("SQL")
            var query = "SELECT * FROM pizza WHERE restaurant_id = ? AND name LIKE ?";
            return database.executeUniqueQuery(Pizza.class, query, stmt -> {
                stmt.setLong(1, restaurantId);
                stmt.setString(2, name + "%");
            }).orElseThrow(() -> new PizzaBotException("Diese Pizza kenne ich nicht."));
        } catch (NonUniqueResultException _) {
            throw new PizzaBotException("Diese Pizza kenne ich nicht.");
        }
    }

    public long order(@NotNull OrderDetail order, @NotNull User user, @NotNull Pizza pizza, @Nullable String notes) throws SQLException {
        return database.transaction(() -> {
            try {
                database.insert(user);
            } catch (SQLException _) {}

            var item = new OrderItem(-1, order.id(), pizza.id(), notes);
            var itemId = database.insert(item);

            var mapping = new OrderItemToUser(itemId, user.id(), false);
            database.insert(mapping);

            return itemId;
        });
    }

    public void add(@NotNull OrderItem item, @NotNull User user, boolean payed) throws SQLException {
        database.insert(new OrderItemToUser(item.id(), user.id(), payed));
    }

    public @NotNull List<@NotNull UserWithPayment> getUsersByItem(@NotNull OrderItem item) throws SQLException {
        @Language("SQL")
        var query = """
            SELECT
                "user"."id" AS "id",
                "user"."username" AS "username",
                "order_item_to_user"."order_item_id" AS "order_item_id",
                "order_item_to_user"."payed" AS "payed"
            FROM "user"
                JOIN "order_item_to_user" ON "user"."id" = "order_item_to_user"."user_id"
            WHERE "order_item_to_user"."order_item_id" = ?
            """;
        return database.executeQuery(UserWithPayment.class, query, stmt -> stmt.setLong(1, item.id()));
    }

    // Revoke

    public @NotNull Optional<OrderItemSummary> getItemById(long itemId) throws SQLException {
        return database.transaction(() -> {
            var item = database.findById(OrderItem.class, itemId);
            if (item.isEmpty()) return Optional.empty();

            var pizza = database.findById(Pizza.class, item.get().pizzaId()).orElseThrow();
            var users = getUsersByItem(item.get());

            return Optional.of(new OrderItemSummary(item.get(), pizza, users));
        });
    }

    public @NotNull Optional<OrderItem> getItemByPizzaName(long orderId, long userId, @NotNull String name) throws SQLException {
        @Language("SQL")
        var query = """
            SELECT "order_item".*
            FROM "order_item"
                JOIN "pizza" ON "order_item"."pizza_id" = "pizza"."id"
                JOIN "order_item_to_user" ON "order_item"."id" = "order_item_to_user"."order_item_id"
            WHERE "order_item"."order_id" = ? AND "order_item_to_user"."user_id" = ? AND "pizza"."name" LIKE ?
            """;
        return database.executeUniqueQuery(OrderItem.class, query, stmt -> {
            stmt.setLong(1, orderId);
            stmt.setLong(2, userId);
            stmt.setString(3, name + "%");
        });
    }

    public @NotNull Optional<OrderItem> getItemByUser(long orderId, long userId) throws SQLException {
        @Language("SQL")
        var query  = """
            SELECT "order_item".*
            FROM "order_item"
                JOIN "order_item_to_user" ON "order_item"."id" = "order_item_to_user"."order_item_id"
            WHERE "order_item"."order_id" = ? AND "order_item_to_user"."user_id" = ?
            """;
        return database.executeUniqueQuery(OrderItem.class, query, stmt -> {
            stmt.setLong(1, orderId);
            stmt.setLong(2, userId);
        });
    }

    public @NotNull List<@NotNull OrderItem> getItemsByUser(long orderId, long userId) throws SQLException {
        @Language("SQL")
        var query = """
                    SELECT "order_item".*
                    FROM "order_item"
                        JOIN "order_item_to_user" ON "order_item"."id" = "order_item_to_user"."order_item_id"
                    WHERE "order_item"."order_id" = ? AND "order_item_to_user"."user_id" = ?
                    """;
        return database.executeQuery(OrderItem.class, query, stmt -> {
            stmt.setLong(1, orderId);
            stmt.setLong(2, userId);
        });
    }

    public boolean delete(@NotNull OrderItem item) throws SQLException {
        return database.delete(OrderItem.class, item.id());
    }

    // Summary

    public @NotNull Order getOrderByEventAndName(long eventId, @NotNull String name) throws SQLException {
        try {
            @Language("SQL")
            var query = """
                SELECT "order".*
                FROM "order"
                WHERE "order"."event_id" = ? AND "order"."name" LIKE ?
                """;

            return database.executeUniqueQuery(Order.class, query, stmt -> {
                stmt.setLong(1, eventId);
                stmt.setString(2, name + "%");
            }).orElseThrow(() -> new PizzaBotException("Diese Bestellung kenne ich nicht."));
        } catch (NonUniqueResultException ex) {
            throw new PizzaBotException("Die Bestellung ist nicht eindeutig.");
        }
    }

    // Order List

    public @NotNull List<@NotNull OrderItemDetail> getItemDetailByEvent(long eventId) throws SQLException {
        return database.findAllWhere(OrderItemDetail.class, OrderItemDetail.Fields.eventId, eventId);
    }

    public @NotNull List<@NotNull OrderItemDetail> getItemDetailByOrder(long orderId) throws SQLException {
        return database.findAllWhere(OrderItemDetail.class, OrderItemDetail.Fields.orderId, orderId);
    }

    public @NotNull List<@NotNull OrderItemDetail> getItemDetailByEventAndUser(long eventId, long userId) throws SQLException {
        @Language("SQL")
        var query = """
            SELECT *
            FROM order_item_detail
            WHERE event_id = ? AND EXISTS(
                SELECT * FROM "order_item_to_user"
                WHERE "user_id" = ? AND "order_item_id" = "order_item_detail"."id"
            )
            """;
        return database.executeQuery(OrderItemDetail.class, query, stmt -> {
            stmt.setLong(1, eventId);
            stmt.setLong(2, userId);
        });
    }
}
