package eu.jonahbauer.chat.server.json;

import com.google.gson.*;
import eu.jonahbauer.chat.bot.config.BotConfig;
import eu.jonahbauer.chat.bot.config.BotConfigurationException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.lang.reflect.Type;

public class BotConfigDeserializer implements JsonDeserializer<BotConfig> {

    @Override
    public @Nullable BotConfig deserialize(@NotNull JsonElement json, @NotNull Type typeOfT, @NotNull JsonDeserializationContext context) throws JsonParseException {
        if (json.isJsonNull()) return null;
        if (!json.isJsonObject()) throw new BotConfigurationException("Expected an object");
        var obj = json.getAsJsonObject();

        var builder = BotConfig.builder();

        obj.asMap().forEach((key, value) -> {
            if (value instanceof JsonPrimitive primitive) {
                if (primitive.isBoolean()) {
                    builder.value(key, primitive.getAsBoolean());
                } else if (primitive.isString()) {
                    builder.value(key, primitive.getAsString());
                } else if (primitive.isNumber()) {
                    switch (primitive.getAsNumber()) {
                        case Long l -> builder.value(key, l);
                        case Double d -> builder.value(key, d);
                        case Number n -> throw new BotConfigurationException("Invalid number: " + n);
                    }
                } else {
                    throw new AssertionError();
                }
            } else if (value.isJsonArray()) {
                var list = value.getAsJsonArray().asList().stream()
                        .map(e -> {
                            if (e instanceof JsonPrimitive primitive && primitive.isString()) {
                                return primitive.getAsString();
                            } else {
                                throw new BotConfigurationException("Unsupported property type in array.");
                            }
                        })
                        .toList();
                builder.value(key, list);
            } else {
                throw new BotConfigurationException("Unsupported property type.");
            }
        });

        return builder.build();
    }

}
