package eu.jonahbauer.chat.server.management.impl;

import eu.jonahbauer.chat.server.management.annotations.ManagedAttribute;
import eu.jonahbauer.chat.server.management.annotations.ManagedBean;
import eu.jonahbauer.chat.server.management.annotations.ManagedOperation;
import eu.jonahbauer.chat.server.management.annotations.ManagedParameter;
import lombok.extern.slf4j.Slf4j;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.management.*;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.function.Function;

@Slf4j
class AdvancedMBean extends StandardMBean {
    public AdvancedMBean(@NotNull Object implementation) {
        super(implementation, null, true);
    }

    @Override
    protected @Nullable String getDescription(@Nullable MBeanInfo info) {
        var clazz = getMBeanInterface();
        var annotation = clazz.getAnnotation(ManagedBean.class);
        if (annotation != null && !"".equals(annotation.description())) {
            return annotation.description();
        } else {
            return super.getDescription(info);
        }
    }

    @Override
    protected int getImpact(@Nullable MBeanOperationInfo info) {
        var method = findOperation(info);
        var annotation = method != null ? method.getAnnotation(ManagedOperation.class) : null;
        if (annotation != null) {
            return annotation.impact().getCode();
        } else {
            return super.getImpact(info);
        }
    }

    @Override
    protected @Nullable String getDescription(@Nullable MBeanOperationInfo info) {
        var method = findOperation(info);
        var annotation = method != null ? method.getAnnotation(ManagedOperation.class) : null;
        if (annotation != null && !"".equals(annotation.description())) {
            return annotation.description();
        } else {
            return super.getDescription(info);
        }
    }

    @Override
    protected @Nullable String getParameterName(@Nullable MBeanOperationInfo op, @Nullable MBeanParameterInfo param, int sequence) {
        var method = findOperation(op);
        var annotation = method != null ? method.getParameters()[sequence].getAnnotation(ManagedParameter.class) : null;
        if (annotation != null && !"".equals(annotation.name())) {
            return annotation.name();
        } else {
            return super.getParameterName(op, param, sequence);
        }
    }

    @Override
    protected @Nullable String getDescription(@Nullable MBeanOperationInfo op, @Nullable MBeanParameterInfo param, int sequence) {
        var method = findOperation(op);
        var annotation = method != null ? method.getParameters()[sequence].getAnnotation(ManagedParameter.class) : null;
        if (annotation != null && !"".equals(annotation.description())) {
            return annotation.description();
        } else {
            return super.getParameterName(op, param, sequence);
        }
    }

    @Override
    protected @Nullable String getDescription(@Nullable MBeanAttributeInfo info) {
        if (info == null) return super.getDescription(info);

        var name = info.getName();
        var type = getType(info, MBeanAttributeInfo::getType);

        var getter = findMethod(("boolean".equals(type) ? "is" : "get") + capitalize(name));
        if (getter != null) {
            var annotation = getter.getAnnotation(ManagedAttribute.class);
            if (annotation != null && !"".equals(annotation.description())) return annotation.description();
        }

        var setter = findMethod("set" + capitalize(name), type);
        if (setter != null) {
            var annotation = setter.getAnnotation(ManagedAttribute.class);
            if (annotation != null && !"".equals(annotation.description())) return annotation.description();
        }

        return super.getDescription(info);
    }

    private <T extends MBeanFeatureInfo> String getType(@NotNull T info, @NotNull Function<T, String> getType) {
        var descriptor = info.getDescriptor();
        if (descriptor == null) return getType.apply(info);

        try {
            var originalType = (String) descriptor.getFieldValue(JMX.ORIGINAL_TYPE_FIELD);
            return originalType.contains("<") ? originalType.substring(0, originalType.indexOf("<")) : originalType;
        } catch (ClassCastException | RuntimeOperationsException ignored) {
            return getType.apply(info);
        }
    }

    private @Nullable Method findOperation(@Nullable MBeanOperationInfo info) {
        if (info == null) return null;

        var name = info.getName();
        var params = new ArrayList<String>();

        for (var parameter : info.getSignature()) {
            params.add(getType(parameter, MBeanParameterInfo::getType));
        }

        return findMethod(name, params);
    }

    private @Nullable Method findMethod(@NotNull String name, @NotNull String @NotNull... parameters) {
        return findMethod(name, Arrays.asList(parameters));
    }

    private @Nullable Method findMethod(@NotNull String name, @NotNull List<@NotNull String> parameters) {
        var clazz = getMBeanInterface();
        methods: for (var method : clazz.getMethods()) {
            if (!method.getName().equals(name)) continue;
            if (method.getParameterCount() != parameters.size()) continue;

            var parameterTypes = method.getParameterTypes();
            for (int i = 0; i < parameters.size(); i++) {
                if (!parameters.get(i).equals(parameterTypes[i].getName())) continue methods;
            }

            return method;
        }
        return null;
    }

    private static @NotNull String capitalize(@NotNull String name) {
        if (name.isEmpty()) return name;
        return name.substring(0, 1).toUpperCase(Locale.ROOT) + name.substring(1);
    }
}
